package uci.inf221.lunarLander.client;

import java.util.UUID;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.log4j.Logger;

import uci.inf221.lunarLander.dataStore.component.AsteroidsComponent;
import uci.inf221.lunarLander.dataStore.component.BulletsComponent;
import uci.inf221.lunarLander.dataStore.component.ClockComponent;
import uci.inf221.lunarLander.dataStore.component.LevelComponent;
import uci.inf221.lunarLander.dataStore.component.ShipsComponent;
import uci.inf221.lunarLander.dataStore.component.StatusComponent;
import uci.inf221.lunarLander.dataStore.connector.DataStoreConnector;
import uci.inf221.lunarLander.gui.component.GUIComponent;
import uci.inf221.lunarLander.logic.component.CollisionDetectionLogicComponent;
import uci.inf221.lunarLander.logic.component.GameLogicComponent;
import uci.inf221.lunarLander.logic.component.PositionLogicComponent;
import uci.inf221.lunarLander.logic.component.ThresholdingLogicComponent;
import uci.inf221.lunarLander.logic.connector.LogicConnector;
import uci.inf221.lunarLander.net.connector.ClientToServerConnector;

public class Client {

	private final static Logger log = Logger.getLogger(Client.class);

	private static Thread thread;
	private static AtomicBoolean terminated = new AtomicBoolean(false);

	private static ClockComponent clockComponent;
	private static StatusComponent statusComponent;
	private static LevelComponent levelComponent;
	private static ShipsComponent shipsComponent;
	private static AsteroidsComponent asteroidsComponent;
	private static BulletsComponent bulletsComponent;
	private static CollisionDetectionLogicComponent collisionDetectionLogicComponent;
	private static PositionLogicComponent positionLogicComponent;
	private static GUIComponent guiComponent;
	private static ThresholdingLogicComponent thresholingLogicComponent;
	private static GameLogicComponent gameLogicComponent;

	private static DataStoreConnector dataStoreConnector;
	private static LogicConnector logicConnector;
	private static ClientToServerConnector clientToServerConnector;

	private static UUID playerId = UUID.randomUUID();

	public static void start() {
		thread = Thread.currentThread();

		log.info("starting up ...");

		Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
			@Override
			public void run() {
				Client.terminate();
			}
		}));

		createStructure();

		startComponents();

		while (!terminated.get()) {
			try {
				Thread.sleep(10);
			} catch (InterruptedException e) {
			}
		}

		log.info("shutting down ...");

		stopComponents();
		waitForComponentsStop();
	}

	private static void createStructure() {
		clockComponent = new ClockComponent();
		statusComponent = new StatusComponent();
		levelComponent = new LevelComponent();
		shipsComponent = new ShipsComponent();
		asteroidsComponent = new AsteroidsComponent();
		bulletsComponent = new BulletsComponent();
		collisionDetectionLogicComponent = new CollisionDetectionLogicComponent();
		positionLogicComponent = new PositionLogicComponent();
		guiComponent = new GUIComponent();
		thresholingLogicComponent = new ThresholdingLogicComponent();
		gameLogicComponent = new GameLogicComponent();

		dataStoreConnector = new DataStoreConnector();
		logicConnector = new LogicConnector();
		clientToServerConnector = new ClientToServerConnector();

		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_CLOCK_COMPONENT)
				.connectTo(
						clockComponent
								.getPort(ClockComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_STATUS_COMPONENT)
				.connectTo(
						statusComponent
								.getPort(StatusComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_LEVEL_COMPONENT)
				.connectTo(
						levelComponent
								.getPort(LevelComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_SHIPS_COMPONENT)
				.connectTo(
						shipsComponent
								.getPort(ShipsComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_ASTEROIDS_COMPONENT)
				.connectTo(
						asteroidsComponent
								.getPort(AsteroidsComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_BULLETS_COMPONENT)
				.connectTo(
						bulletsComponent
								.getPort(BulletsComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_LOGIC_CONNECTOR)
				.connectTo(
						logicConnector
								.getPort(LogicConnector.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(
						DataStoreConnector.PORT_IDENTIFIER_COLLISION_DETECTION_LOGIC_COMPONENT)
				.connectTo(
						collisionDetectionLogicComponent
								.getPort(CollisionDetectionLogicComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(
						DataStoreConnector.PORT_IDENTIFIER_POSITION_LOGIC_COMPONENT)
				.connectTo(
						positionLogicComponent
								.getPort(PositionLogicComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_GUI_COMPONENT)
				.connectTo(
						guiComponent
								.getPort(GUIComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(
						DataStoreConnector.PORT_IDENTIFIER_CLIENT_TO_SERVER_CONNECTOR)
				.connectTo(
						clientToServerConnector
								.getPort(ClientToServerConnector.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(
						DataStoreConnector.PORT_IDENTIFIER_THRESHOLDING_LOGIC_COMPONENT)
				.connectTo(
						thresholingLogicComponent
								.getPort(ThresholdingLogicComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));

		logicConnector
				.getPort(
						LogicConnector.PORT_IDENTIFIER_COLLISION_DETECTION_LOGIC_COMPONENT)
				.connectTo(
						collisionDetectionLogicComponent
								.getPort(CollisionDetectionLogicComponent.PORT_IDENTIFIER_LOGIC_CONNECTOR));
		logicConnector
				.getPort(LogicConnector.PORT_IDENTIFIER_GAME_LOGIC_COMPONENT)
				.connectTo(
						gameLogicComponent
								.getPort(GameLogicComponent.PORT_IDENTIFIER_LOGIC_CONNECTOR));

		clientToServerConnector
				.getPort(
						ClientToServerConnector.PORT_IDENTIFIER_THRESHOLDING_LOGIC_COMPONENT)
				.connectTo(
						thresholingLogicComponent
								.getPort(ThresholdingLogicComponent.PORT_IDENTIFIER_CLIENT_TO_SERVER_CONNECTOR));
	}

	private static void startComponents() {
		clockComponent.start();
		statusComponent.start();
		levelComponent.start();
		shipsComponent.start();
		asteroidsComponent.start();
		bulletsComponent.start();
		positionLogicComponent.start();
		collisionDetectionLogicComponent.start();
		gameLogicComponent.start();
		guiComponent.start();
		thresholingLogicComponent.start();

		dataStoreConnector.start();
		logicConnector.start();
		clientToServerConnector.start();
	}

	private static void stopComponents() {
		thresholingLogicComponent.stop();
		guiComponent.stop();
		gameLogicComponent.stop();
		collisionDetectionLogicComponent.stop();
		positionLogicComponent.stop();
		bulletsComponent.stop();
		asteroidsComponent.stop();
		shipsComponent.stop();
		levelComponent.stop();
		statusComponent.stop();
		clockComponent.stop();

		clientToServerConnector.stop();
		logicConnector.stop();
		dataStoreConnector.stop();
	}

	public static void waitForComponentsStop() {
		thresholingLogicComponent.waitForStop();
		guiComponent.waitForStop();
		gameLogicComponent.waitForStop();
		collisionDetectionLogicComponent.waitForStop();
		positionLogicComponent.waitForStop();
		bulletsComponent.waitForStop();
		asteroidsComponent.waitForStop();
		shipsComponent.waitForStop();
		levelComponent.waitForStop();
		statusComponent.waitForStop();
		clockComponent.waitForStop();

		clientToServerConnector.waitForStop();
		logicConnector.waitForStop();
		dataStoreConnector.waitForStop();
	}

	public static void terminate() {
		terminated.set(true);
		thread.interrupt();
	}

	public static UUID getPlayerId() {
		return playerId;
	}

}
