package uci.inf221.lunarLander.logic.component;

import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.log4j.Logger;

import uci.inf221.lunarLander.arch.c2.C2Component;
import uci.inf221.lunarLander.arch.c2.C2Notification;
import uci.inf221.lunarLander.arch.c2.C2Request;
import uci.inf221.lunarLander.client.Client;
import uci.inf221.lunarLander.dataStore.entities.Asteroid;
import uci.inf221.lunarLander.dataStore.entities.Bullet;
import uci.inf221.lunarLander.dataStore.entities.Level.LevelStatus;
import uci.inf221.lunarLander.dataStore.entities.MoveableDataEntity;
import uci.inf221.lunarLander.dataStore.entities.MovementVector;
import uci.inf221.lunarLander.dataStore.entities.Ship;
import uci.inf221.lunarLander.dataStore.message.CreateNewShipRequest;
import uci.inf221.lunarLander.dataStore.message.DestroyAllEntitiesRequest;
import uci.inf221.lunarLander.dataStore.message.DestroyEntityRequest;
import uci.inf221.lunarLander.dataStore.message.EntityDestroyedNotification;
import uci.inf221.lunarLander.dataStore.message.LevelStatusUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.MovementVectorUpdatedNotification;
import uci.inf221.lunarLander.dataStore.message.NewAsteroidNotification;
import uci.inf221.lunarLander.dataStore.message.NewBulletNotification;
import uci.inf221.lunarLander.dataStore.message.NewShipNotification;
import uci.inf221.lunarLander.dataStore.message.UpdateGameStartTimeRequest;
import uci.inf221.lunarLander.dataStore.message.UpdateLevelStatusRequest;
import uci.inf221.lunarLander.logic.message.EntityCollisionNotification;
import uci.inf221.lunarLander.logic.message.EntityLevelCollisionNotification;
import uci.inf221.lunarLander.logic.message.EntityOutsideMapNotification;
import uci.inf221.lunarLander.logic.message.ShipTouchingNotification;

public class GameLogicComponent extends C2Component {

	public static final String PORT_IDENTIFIER_LOGIC_CONNECTOR = "logic";

	private final Logger log = Logger.getLogger(this.getClass());

	// TODO adjust constants ?
	// 10 degrees in both directions
	private static final double SHIP_TOUCHING_ANGLE_ALLOWANCE = Math.PI / 18;
	// px per second
	private static final float SHIP_TOUCHING_MAX_SPEED = 6f / 1000;

	private Map<UUID, MoveableDataEntity> moveableEntities = new ConcurrentHashMap<UUID, MoveableDataEntity>();
	private AtomicInteger ships = new AtomicInteger(0);
	private AtomicBoolean playerShipCreated = new AtomicBoolean(false);

	@Override
	protected void createPorts() {
		addTopPort(PORT_IDENTIFIER_LOGIC_CONNECTOR);
	}

	@Override
	public void start() {
	}

	@Override
	protected void initialize() {
	}

	@Override
	public void stop() {
	}

	@Override
	public void waitForStop() {
	}

	@Override
	protected void receiveNotification(C2Notification notification) {
		if (notification instanceof ShipTouchingNotification) {
			UUID entityId = ((ShipTouchingNotification) notification)
					.getEntityId();
			Ship ship = (Ship) moveableEntities.get(entityId);
			if (ship != null && ship.getPlayerId().equals(Client.getPlayerId())) {
				float accelerationDir = ship.getMovementVector()
						.getAccelerationDirection();
				float speedMag = ship.getMovementVector().getSpeedMagnitude();

				// check angle of acceleration and speed magnitude
				if (Math.abs(accelerationDir + Math.PI / 2) < SHIP_TOUCHING_ANGLE_ALLOWANCE
						&& speedMag < SHIP_TOUCHING_MAX_SPEED) {
					sendRequest(new UpdateLevelStatusRequest(this,
							LevelStatus.COMPLETED));
				} else {
					sendRequest(new DestroyEntityRequest(this, entityId));
				}
			} // otherwise will be handled by remote client
		} else if (notification instanceof EntityLevelCollisionNotification) {
			UUID entityId = ((EntityLevelCollisionNotification) notification)
					.getEntityId();
			MoveableDataEntity entity = moveableEntities.get(entityId);
			if (entity != null) {
				if (!(entity instanceof Ship)
						|| ((Ship) entity).getPlayerId().equals(
								Client.getPlayerId())) {
					sendRequest(new DestroyEntityRequest(this, entityId));
				} // otherwise will be handled by remote client
			}
		} else if (notification instanceof EntityCollisionNotification) {
			UUID entityId1 = ((EntityCollisionNotification) notification)
					.getEntityId1();
			UUID entityId2 = ((EntityCollisionNotification) notification)
					.getEntityId2();
			MoveableDataEntity entity1 = moveableEntities.get(entityId1);
			MoveableDataEntity entity2 = moveableEntities.get(entityId2);

			if (entity1 == null || entity2 == null)
				return;

			boolean handle = true;

			if (entity1 instanceof Ship) {
				if (entity2 instanceof Ship) {
					// handle = ((Ship) entity1).getPlayerId().equals(
					// Client.getPlayerId())
					// || ((Ship) entity2).getPlayerId().equals(
					// Client.getPlayerId());
					handle = false; // ships cannot collide with each other
				} else {
					handle = ((Ship) entity1).getPlayerId().equals(
							Client.getPlayerId());
				}
			} else if (entity2 instanceof Ship) {
				handle = ((Ship) entity2).getPlayerId().equals(
						Client.getPlayerId());
			}

			if (handle) {
				if (!(entity1 instanceof Bullet)) {
					if (!(entity2 instanceof Bullet)
							|| !((Bullet) entity2).getOriginEntityId().equals(
									entity1.getId())) {
						sendRequest(new DestroyEntityRequest(this, entityId1));
					}
				}
				if (!(entity2 instanceof Bullet)) {
					if (!(entity1 instanceof Bullet)
							|| !((Bullet) entity1).getOriginEntityId().equals(
									entity2.getId())) {
						sendRequest(new DestroyEntityRequest(this, entityId2));
					}
				}
			}
		} else if (notification instanceof EntityOutsideMapNotification) {
			UUID entityId = ((EntityOutsideMapNotification) notification)
					.getEntityId();
			MoveableDataEntity entity = moveableEntities.get(entityId);
			if (entity != null) {
				if (!(entity instanceof Ship)
						|| ((Ship) entity).getPlayerId().equals(
								Client.getPlayerId())) {
					sendRequest(new DestroyEntityRequest(this, entityId));
				} // otherwise will be handled by remote client
			}
		} else if (notification instanceof MovementVectorUpdatedNotification) {
			MoveableDataEntity entity = moveableEntities
					.get(((MovementVectorUpdatedNotification) notification)
							.getEntityId());
			if (entity != null) {
				synchronized (entity) {
					entity.setMovementVector(((MovementVectorUpdatedNotification) notification)
							.getMovementVector());
				}
			} else {
				log.warn("movement vector update for unkown entity of id: "
						+ ((MovementVectorUpdatedNotification) notification)
								.getEntityId());
			}
		} else if (notification instanceof NewAsteroidNotification) {
			Asteroid asteroid = ((NewAsteroidNotification) notification)
					.getAsteroid();
			moveableEntities.put(asteroid.getId(), asteroid);
		} else if (notification instanceof NewBulletNotification) {
			Bullet bullet = ((NewBulletNotification) notification).getBullet();
			moveableEntities.put(bullet.getId(), bullet);
		} else if (notification instanceof NewShipNotification) {
			Ship ship = ((NewShipNotification) notification).getShip();
			if (moveableEntities.put(ship.getId(), ship) == null) {
				ships.incrementAndGet();
			}
		} else if (notification instanceof EntityDestroyedNotification) {
			MoveableDataEntity entity = moveableEntities
					.remove(((EntityDestroyedNotification) notification)
							.getEntityId());

			if (entity != null && entity instanceof Ship) {
				if (ships.decrementAndGet() == 0) {
					sendRequest(new UpdateLevelStatusRequest(this,
							LevelStatus.FAILED));
				}
			}
		} else if (notification instanceof LevelStatusUpdatedNotification) {
			LevelStatus levelStatus = ((LevelStatusUpdatedNotification) notification)
					.getLevelStatus();
			if (levelStatus.hasEnded()) {
				moveableEntities.clear();
				ships.set(0);
				playerShipCreated.set(false);
				sendRequest(new UpdateGameStartTimeRequest(this, 0));
				sendRequest(new DestroyAllEntitiesRequest(this));
			} else if (levelStatus == LevelStatus.LOADED
					|| levelStatus == LevelStatus.STARTED) {
				// create own ship
				if (!playerShipCreated.getAndSet(true)) {
					Ship ship = new Ship();
					ship.setPlayerId(Client.getPlayerId());
					MovementVector movementVector = new MovementVector();
					movementVector.setPosition(new float[] { ship.getExtent(),
							ship.getExtent() });
					movementVector.setAcceleration(0, 0);
					movementVector.setSpeed(Ship.INITIAL_SPEED);
					ship.setMovementVector(movementVector);
					sendRequest(new CreateNewShipRequest(this, ship));
				}
			}
		}
	}

	@Override
	protected void receiveRequest(C2Request request) {
	}

}
