package uci.inf221.lunarLander.net.connector.serverToClients.component;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;

import org.apache.log4j.Logger;
import org.cynetec.lib.system.net.ntp.NTPUDPServer;

import uci.inf221.lunarLander.arch.Message;
import uci.inf221.lunarLander.arch.c2.C2Component;
import uci.inf221.lunarLander.arch.c2.C2Notification;
import uci.inf221.lunarLander.arch.c2.C2Request;
import uci.inf221.lunarLander.net.connector.serverToClients.message.HandleConnectionStateChangeRequest;
import uci.inf221.lunarLander.net.connector.serverToClients.message.HandleIncomingMessageRequest;
import uci.inf221.lunarLander.net.connector.serverToClients.message.HandleServerStartedRequest;
import uci.inf221.lunarLander.net.connector.serverToClients.message.NewMessageNotification;
import uci.inf221.lunarLander.net.connector.serverToClients.message.NewMessageNotification.TargetMode;
import uci.inf221.lunarLander.server.Server;
import uci.inf221.lunarLander.utils.KryoInitializer;

import com.esotericsoftware.kryonet.Connection;
import com.esotericsoftware.kryonet.Listener;
import com.esotericsoftware.kryonet.Listener.ThreadedListener;

public class ClientConnectionManagerComponent extends C2Component {

	public class ServerListener extends ThreadedListener {

		public ServerListener() {
			super(new Listener() {

				@Override
				public void connected(Connection connection) {
					sendRequest(new HandleConnectionStateChangeRequest(
							ClientConnectionManagerComponent.this,
							connection.getID(), true));
					log.info("client connected: " + connection.getID());
				}

				@Override
				public void disconnected(Connection connection) {
					sendRequest(new HandleConnectionStateChangeRequest(
							ClientConnectionManagerComponent.this,
							connection.getID(), false));
					log.info("client disconnected: " + connection.getID());
				}

				@Override
				public void received(Connection connection, Object object) {
					if (object instanceof Message) {
						sendRequest(new HandleIncomingMessageRequest(
								ClientConnectionManagerComponent.this,
								(Message) object, connection.getID()));
						log.trace("new message from client "
								+ connection.getID() + ": " + object.toString());
					} else {
						log.warn("discarding unknown message from client "
								+ connection.getID() + ": " + object.toString());
					}
				}

			});
		}

		public void terminate() {
			threadPool.shutdown();
		}

		public boolean isTerminated() {
			return threadPool.isTerminated();
		}

	}

	public static final String PORT_IDENTIFIER_CONNECTION_MANAGER_CONNECTOR = "connection_manager";

	private static final int WRITE_BUFFER_SIZE = 1 << 16;
	private static final int OBJECT_BUFFER_SIZE = 1 << 15;

	private Logger log = Logger.getLogger(this.getClass());

	private com.esotericsoftware.kryonet.Server server;
	private NTPUDPServer ntpServer;
	private ServerListener serverListener;

	@Override
	protected void createPorts() {
		addTopPort(PORT_IDENTIFIER_CONNECTION_MANAGER_CONNECTOR);
	}

	@Override
	protected void initialize() {
		server = new com.esotericsoftware.kryonet.Server(WRITE_BUFFER_SIZE,
				OBJECT_BUFFER_SIZE);
		KryoInitializer.initKryo(server.getKryo());

		server.addListener((serverListener = new ServerListener()));

		// init ntp server
		try {
			ntpServer = new NTPUDPServer(Server.NTP_PORT);
		} catch (IOException e) {
			log.error("could not initialize NTP server");
		}
	}

	@Override
	public void start() {
		server.start();
		log.info("started server");
		ntpServer.start();
		log.info("started ntpServer");

		try {
			sendRequest(new HandleServerStartedRequest(this, InetAddress
					.getLocalHost().getHostAddress()));
		} catch (UnknownHostException e) {
			throw new RuntimeException("no local host available", e);
		}

		try {
			server.bind(Server.TCP_PORT, Server.UDP_PORT);
		} catch (IOException e) {
			throw new RuntimeException("server could not bind to ports", e);
		}
	}

	@Override
	public void stop() {
		ntpServer.terminate();
		server.stop();
	}

	@Override
	public void waitForStop() {
		while (ntpServer.isAlive()) {
			try {
				Thread.sleep(1);
			} catch (InterruptedException e) {
			}
		}

		serverListener.terminate();
		while (!serverListener.isTerminated()) {
			try {
				Thread.sleep(1);
			} catch (InterruptedException e) {
			}
		}
	}

	@Override
	protected void receiveNotification(C2Notification notification) {
		if (notification instanceof NewMessageNotification) {
			NewMessageNotification newMessageNotification = (NewMessageNotification) notification;
			Integer targetConnectionId = newMessageNotification
					.getTargetConnectionId();
			Message message = newMessageNotification.getMessage();
			if (targetConnectionId != null) {
				if (newMessageNotification.getTargetMode() == TargetMode.SINGLE) {
					if (newMessageNotification.isImportant()) {
						log.trace("sending message to client "
								+ targetConnectionId + " (TCP): "
								+ message.toString());
						server.sendToTCP(targetConnectionId, message);
					} else {
						log.trace("sending message to client "
								+ targetConnectionId + " (UDP): "
								+ message.toString());
						server.sendToUDP(targetConnectionId, message);
					}
				} else {
					if (newMessageNotification.isImportant()) {
						log.trace("sending message to all clients except "
								+ targetConnectionId + " (TCP): "
								+ message.toString());
					} else {
						log.trace("sending message to all clients except "
								+ targetConnectionId + " (UDP): "
								+ message.toString());

					}

					Connection[] connections = server.getConnections();
					for (int i = 0, n = connections.length; i < n; i++) {
						Connection connection = connections[i];
						if (connection.getID() != targetConnectionId) {
							if (newMessageNotification.isImportant()) {
								server.sendToTCP(connection.getID(), message);
							} else {
								server.sendToUDP(connection.getID(), message);
							}
						}
					}
				}
			} else {
				if (newMessageNotification.isImportant()) {
					log.trace("sending message to all clients (TCP): "
							+ message.toString());
					server.sendToAllTCP(message);
				} else {
					log.trace("sending message to all clients (UDP): "
							+ message.toString());
					server.sendToAllUDP(message);
				}
			}
		}
	}

	@Override
	protected void receiveRequest(C2Request request) {
	}
}
