package uci.inf221.lunarLander.server;

import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.log4j.Logger;

import uci.inf221.lunarLander.dataStore.component.AsteroidsComponent;
import uci.inf221.lunarLander.dataStore.component.BulletsComponent;
import uci.inf221.lunarLander.dataStore.component.ClockComponent;
import uci.inf221.lunarLander.dataStore.component.LevelComponent;
import uci.inf221.lunarLander.dataStore.component.ShipsComponent;
import uci.inf221.lunarLander.dataStore.component.StatusComponent;
import uci.inf221.lunarLander.dataStore.connector.DataStoreConnector;
import uci.inf221.lunarLander.managing.component.AsteroidManagerComponent;
import uci.inf221.lunarLander.managing.component.LevelManagerComponent;
import uci.inf221.lunarLander.net.connector.ServerToClientsConnector;

public class Server {

	public static final int TCP_PORT = 60432;
	public static final int UDP_PORT = 60433;
	public static final int NTP_PORT = 60434;

	private final static Logger log = Logger.getLogger(Server.class);
	public static final int MIN_PLAYERS = 1;

	private static Thread thread;
	private static AtomicBoolean terminated = new AtomicBoolean(false);

	private static ClockComponent clockComponent;
	private static StatusComponent statusComponent;
	private static LevelComponent levelComponent;
	private static ShipsComponent shipsComponent;
	private static AsteroidsComponent asteroidsComponent;
	private static BulletsComponent bulletsComponent;
	private static AsteroidManagerComponent asteroidManagerComponent;
	private static LevelManagerComponent bulletManagerComponent;

	private static DataStoreConnector dataStoreConnector;
	private static ServerToClientsConnector serverToClientsConnector;

	public static void start() {
		thread = Thread.currentThread();

		log.info("starting up ...");

		Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
			@Override
			public void run() {
				Server.terminate();
			}
		}));

		createStructure();

		startComponents();

		while (!terminated.get()) {
			try {
				Thread.sleep(10);
			} catch (InterruptedException e) {
			}
		}

		log.info("shutting down ...");

		stopComponents();
		waitForComponentsStop();
	}

	private static void createStructure() {
		clockComponent = new ClockComponent();
		statusComponent = new StatusComponent();
		levelComponent = new LevelComponent();
		shipsComponent = new ShipsComponent();
		asteroidsComponent = new AsteroidsComponent();
		bulletsComponent = new BulletsComponent();
		asteroidManagerComponent = new AsteroidManagerComponent();
		bulletManagerComponent = new LevelManagerComponent();

		dataStoreConnector = new DataStoreConnector();
		serverToClientsConnector = new ServerToClientsConnector();

		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_CLOCK_COMPONENT)
				.connectTo(
						clockComponent
								.getPort(ClockComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_STATUS_COMPONENT)
				.connectTo(
						statusComponent
								.getPort(StatusComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_LEVEL_COMPONENT)
				.connectTo(
						levelComponent
								.getPort(LevelComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_SHIPS_COMPONENT)
				.connectTo(
						shipsComponent
								.getPort(ShipsComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_ASTEROIDS_COMPONENT)
				.connectTo(
						asteroidsComponent
								.getPort(AsteroidsComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(DataStoreConnector.PORT_IDENTIFIER_BULLETS_COMPONENT)
				.connectTo(
						bulletsComponent
								.getPort(BulletsComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(
						DataStoreConnector.PORT_IDENTIFIER_ASTEROID_MANAGER_COMPONENT)
				.connectTo(
						asteroidManagerComponent
								.getPort(AsteroidManagerComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(
						DataStoreConnector.PORT_IDENTIFIER_LEVEL_MANAGER_COMPONENT)
				.connectTo(
						bulletManagerComponent
								.getPort(LevelManagerComponent.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
		dataStoreConnector
				.getPort(
						DataStoreConnector.PORT_IDENTIFIER_SERVER_TO_CLIENTS_CONNECTOR)
				.connectTo(
						serverToClientsConnector
								.getPort(ServerToClientsConnector.PORT_IDENTIFIER_DATA_STORE_CONNECTOR));
	}

	private static void startComponents() {
		clockComponent.start();
		statusComponent.start();
		levelComponent.start();
		shipsComponent.start();
		asteroidsComponent.start();
		bulletsComponent.start();
		asteroidManagerComponent.start();
		bulletManagerComponent.start();

		dataStoreConnector.start();
		serverToClientsConnector.start();
	}

	private static void stopComponents() {
		bulletManagerComponent.stop();
		asteroidManagerComponent.stop();
		bulletsComponent.stop();
		asteroidsComponent.stop();
		shipsComponent.stop();
		levelComponent.stop();
		statusComponent.stop();
		clockComponent.stop();

		serverToClientsConnector.stop();
		dataStoreConnector.stop();
	}

	public static void waitForComponentsStop() {
		bulletManagerComponent.waitForStop();
		asteroidManagerComponent.waitForStop();
		bulletsComponent.waitForStop();
		asteroidsComponent.waitForStop();
		shipsComponent.waitForStop();
		levelComponent.waitForStop();
		statusComponent.waitForStop();
		clockComponent.waitForStop();

		serverToClientsConnector.waitForStop();
		dataStoreConnector.waitForStop();
	}

	public static void terminate() {
		terminated.set(true);
		thread.interrupt();
	}

}
